import asyncio
from typing import Dict

from .scraper_helper import ScrapeMapper
from ..base_functions.base_functions import log
from ..base_functions.data_classes import AuthData, SkipData
from ..client.client import Client


async def scrape(urls, client: Client, file_args: Dict, jdownloader_args: Dict, runtime_args: Dict,
                 jdownloader_auth: AuthData, simpcity_auth: AuthData, socialmediagirls_auth: AuthData,
                 xbunker_auth: AuthData, skip_data: SkipData, quiet=False, close=True):
    await log("Starting Scrape", quiet=quiet)

    scraper = ScrapeMapper(client=client, file_args=file_args, jdownloader_args=jdownloader_args,
                           runtime_args=runtime_args, xbunker_auth=xbunker_auth,
                           socialmediagirls_auth=socialmediagirls_auth, simpcity_auth=simpcity_auth,
                           jdownloader_auth=jdownloader_auth, skip_data=skip_data, quiet=quiet)
    tasks = []
    for link in urls:
        tasks.append(scraper.map_url(link))
    await asyncio.gather(*tasks)

    if close:
        await scraper.close()

    Cascade = scraper.Cascade
    await Cascade.dedupe()
    return Cascade



from typing import Dict

import aiofiles
import logging
import os
import re
from pathlib import Path

import psutil as psutil
import yaml
from colorama import Fore, Style
from yarl import URL

from .config_schema import config_default, files_args, authentication_args, jdownloader_args, runtime_args

"""This file contains generic information and functions that are used around the program"""

FILE_FORMATS = {
    'Images': {
        '.jpg', '.jpeg', '.png', '.gif',
        '.webp', '.jpe', '.svg', '.jfif',
        '.tif', '.tiff', '.jif',
    },
    'Videos': {
        '.mpeg', '.avchd', '.webm', '.mpv',
        '.swf', '.avi', '.m4p', '.wmv',
        '.mp2', '.m4v', '.qt', '.mpe',
        '.mp4', '.flv', '.mov', '.mpg',
        '.ogg', '.mkv', '.mts', '.ts',
    },
    'Audio': {
        '.mp3', '.flac', '.wav', '.m4a',
    },
    'Other': {
        '.json', '.torrent', '.zip', '.rar',
        '.7z', '.torrent', '.psd', '.pdf',
    }
}

MAX_FILENAME_LENGTH = 100

logger = logging.getLogger(__name__)


class FailureException(Exception):
    """Basic failure exception I can throw to force a retry."""

    def __init__(self, code, message="Something went wrong", rescrape=False):
        self.code = code
        self.message = message
        self.rescrape = rescrape
        super().__init__(self.message)
        super().__init__(self.code)


async def sanitize(name: str) -> str:
    return re.sub(r'[<>:"/\\|?*\']', "", name).strip()


async def log(text, style=Fore.WHITE, quiet=False) -> None:
    logger.debug(text)
    if not quiet:
        print(style + str(text) + Style.RESET_ALL)


async def clear() -> None:
    """Clears the terminal screen"""
    os.system('cls' if os.name == 'nt' else 'clear')


async def make_title_safe(title: str):
    title = re.sub(r'[\\*?:"<>|./]', "-", title)
    return title


async def purge_dir(dirname, in_place=True):
    deleted = []
    dir_tree = list(os.walk(dirname, topdown=False))

    for tree_element in dir_tree:
        sub_dir = tree_element[0]
        dir_count = len(os.listdir(sub_dir))
        if dir_count == 0:  # Helps with readability and i've had issues with it deleting non-empty dirs
            deleted.append(sub_dir)

    if in_place:
        list(map(os.rmdir, deleted))


async def regex_links(urls) -> list:
    yarl_links = []
    for line in urls:
        if line.lstrip().startswith('#'):
            continue

        all_links = [x.group().replace(".md.", ".") for x in re.finditer(
            r"(?:http.*?)(?=($|\n|\r\n|\r|\s|\"|\[/URL]|]\[|\[/img]))", line)]
        for link in all_links:
            yarl_links.append(URL(link))
    return yarl_links


async def check_free_space(required_space: int, download_directory: Path):
    free_space = psutil.disk_usage(str(download_directory.parent)).free
    free_space = ((free_space / 1024) / 1024) / 1024
    if required_space > free_space:
        return False
    else:
        return True


async def cyberdrop_parse(url: URL) -> URL:
    mapping_direct = [r'img-...cyberdrop...',
                      r'f.cyberdrop...', r'fs-...cyberdrop...']
    url = str(url)
    for mapping in mapping_direct:
        url = re.sub(mapping, 'cyberdrop.to', url)
    return URL(url)


async def write_last_post_file(file: Path, url: str):
    async with aiofiles.open(file, mode='a') as f:
        await f.write(url + '\n')
    return


async def check_direct(url: URL):
    mapping_direct = [r'i.pixl.li', r'i..pixl.li', r'img-...cyberdrop...', r'f.cyberdrop...',
                      r'fs-...cyberdrop...', r'i.bunkr...', r'i..bunkr...', r'i...bunkr...', r'media-files.bunkr...',
                      r'media-files..bunkr...', r'cdn.bunkr...', r'cdn..bunkr...', r'cdn...bunkr...',
                      r'jpg.church/images/...', r'simp..jpg.church', r's..putmega.com', r's..putme.ga', r'images..imgbox.com' ]
    return any(re.search(domain, url.host) for domain in mapping_direct)


async def is_forum(url: URL):
    mapping_forum = ['simpcity...', 'socialmediagirls....', 'xbunker...']
    return any(re.search(domain, url.host) for domain in mapping_forum)

async def create_config(config: Path, passed_args=None, remake=None):
    if config.is_file() and not remake:
        logging.debug("Validating Config")
        await validate_config(config)
        return

    logging.debug("Creating Config File")
    config_data = config_default
    if passed_args:
        for arg in authentication_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Authentication"][arg] = passed_args[arg]
        for arg in files_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Files"][arg] = str(passed_args[arg])
        for arg in jdownloader_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["JDownloader"][arg] = passed_args[arg]
        for arg in runtime_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Runtime"][arg] = passed_args[arg]

    with open(config, 'w') as yamlfile:
        yaml.dump(config_data, yamlfile)
    return


async def validate_config(config: Path):
    with open(config, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    data = data[0]["Configuration"]
    recreate = 0
    try:
        if not set(authentication_args).issubset(set(data['Authentication'].keys())):
            recreate = 1
        if not set(files_args).issubset(set(data['Files'].keys())):
            recreate = 1
        if not set(jdownloader_args).issubset(set(data['JDownloader'].keys())):
            recreate = 1
        if not set(runtime_args).issubset(set(data['Runtime'].keys())):
            recreate = 1

        if recreate:
            config.unlink()
            logging.debug("Recreating Config")

            args = {}
            args_list = [data['Authentication'], data['Files'], data['JDownloader'], data['Runtime']]
            for dic in args_list:
                args.update(dic)
            await create_config(config, args, True)

    except KeyError:
        config.unlink()
        await log("Config was malformed, recreating.")
        await create_config(config, None, True)


async def run_args(config: Path, cmd_arg: Dict):
    with open(config, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    data = data[0]["Configuration"]
    if data['Apply_Config']:
        return data

    logging.debug("Gathering Args")
    config_data = config_default[0]["Configuration"]
    for arg in authentication_args:
        if arg in cmd_arg.keys():
            config_data["Authentication"][arg] = cmd_arg[arg]
    for arg in files_args:
        if arg in cmd_arg.keys():
            config_data["Files"][arg] = cmd_arg[arg]
    for arg in jdownloader_args:
        if arg in cmd_arg.keys():
            config_data["JDownloader"][arg] = cmd_arg[arg]
    for arg in runtime_args:
        if arg in cmd_arg.keys():
            config_data["Runtime"][arg] = cmd_arg[arg]
    return config_data



from typing import Dict

from yarl import URL

from ..base_functions.base_functions import log, logger, check_direct
from ..base_functions.data_classes import DomainItem
from ..client.client import Session


class ImgBoxCrawler:
    def __init__(self, *, include_id=False, quiet: bool):
        self.include_id = include_id
        self.quiet = quiet

    async def fetch(self, session: Session, url: URL):
        domain_obj = DomainItem(url.host, {})
        await log("Starting scrape of " + str(url), quiet=self.quiet)

        if await check_direct(url):
            await domain_obj.add_to_album(link=URL(url), referral=url, title="ImgBox Loose Files")
            return domain_obj

        try:
            if "g" in url.parts:
                images = await self.folder(session, url)
                for img in images:
                    await domain_obj.add_to_album(url.raw_name, img, url)
            else:
                img = await self.singular(session, url)
                await domain_obj.add_to_album("Loose ImgBox Files", img, url)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url), quiet=self.quiet)
            logger.debug(e)

        await log("Finished scrape of " + str(url), quiet=self.quiet)

        return domain_obj

    async def folder(self, session: Session, url: URL):
        soup = await session.get_BS4(url)
        output = []
        
        images = soup.find('div', attrs={'id': 'gallery-view-content'})
        for link in images.findAll("img"):
            output.append(URL(link.get('src').replace("thumbs", "images").replace("_b", "_o")))

        return output

    async def singular(self, session: Session, url: URL):
        soup = await session.get_BS4(url)
        link = URL(soup.select_one("img[id=img]").get('src'))
        return link



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe
from ..base_functions.data_classes import DomainItem
from ..client.client import Session


class FapelloCrawler:
    def __init__(self, *, include_id=False, quiet: bool):
        self.include_id = include_id
        self.quiet = quiet

    async def fetch(self, session: Session, url: URL):
        await log("Starting scrape of " + str(url), quiet=self.quiet)
        domain_obj = DomainItem('fapello', {})
        results = []

        results.extend(await self.parse_profile(session, url))

        for result in results:
            await domain_obj.add_to_album(result[0], result[1], result[2])

        await log("Finished scrape of " + str(url), quiet=self.quiet)
        return domain_obj

    async def parse_profile(self, session: Session, url: URL):
        try:
            soup, returned_url = await session.get_BS4_and_url(url)
            title = soup.select_one('h2[class="font-semibold lg:text-2xl text-lg mb-2 mt-4"]').get_text()
            title = title + " (Fapello)"
            results = []

            # Fapello does circular looped paging
            if returned_url != url:
                return results

            content_section = soup.select_one("div[id=content]")
            posts = content_section.select("a")
            for post in posts:
                link = post.get('href')
                if link:
                    results.extend(await self.parse_post(session, link, title))

            next_page = soup.select_one('div[id="next_page"] a')
            if next_page:
                next_page = next_page.get('href')
                if next_page:
                    results.extend(await self.parse_profile(session, URL(next_page)))
            return results

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url), quiet=self.quiet)
            logger.debug(e)
            return []

    async def parse_post(self, session: Session, url: URL, title=None):
        try:
            await log("Scraping post: " + str(url), quiet=self.quiet)
            soup = await session.get_BS4(url)
            results = []

            content_section = soup.select_one('div[class="flex justify-between items-center"]')

            images = content_section.select("img")
            for image in images:
                download_link = URL(image.get('src'))
                results.append([title, download_link, url])

            videos = content_section.select("source")
            for video in videos:
                download_link = URL(video.get('src'))
                results.append([title, download_link, url])

            return results

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url), quiet=self.quiet)
            logger.debug(e)
            return []